@@ -1,6 +1,9 @@
 #' @name readRDS.lgb.Booster
-#' @title readRDS for \code{lgb.Booster} models
-#' @description Attempts to load a model stored in a \code{.rds} file, using \code{\link[base]{readRDS}}
+#' @title readRDS for \code{lgb.Booster} models (DEPRECATED)
+#' @description Calls \code{readRDS} in what is expected to be a serialized \code{lgb.Booster} object,
+#'              and then restores its handle through \code{lgb.restore_handle}.
+#'
+#'              \bold{This function throws a warning and will be removed in future versions.}
 #' @param file a connection or the name of the file where the R object is saved to or read from.
 #' @param refhook a hook function for handling reference objects.
 #'
@@ -9,6 +12,8 @@
 #' @examples
 #' \donttest{
 #' library(lightgbm)
+#' \dontshow{setLGBMthreads(2L)}
+#' \dontshow{data.table::setDTthreads(1L)}
 #' data(agaricus.train, package = "lightgbm")
 #' train <- agaricus.train
 #' dtrain <- lgb.Dataset(train$data, label = train$label)
@@ -20,6 +25,7 @@
 #'   , metric = "l2"
 #'   , min_data = 1L
 #'   , learning_rate = 1.0
+#'   , num_threads = 2L
 #' )
 #' valids <- list(test = dtest)
 #' model <- lgb.train(
@@ -36,27 +42,9 @@
 #' @export
 readRDS.lgb.Booster <- function(file, refhook = NULL) {
 
+  warning("'readRDS.lgb.Booster' is deprecated and will be removed in a future release. Use readRDS() instead.")
+
   object <- readRDS(file = file, refhook = refhook)
-
-  # Check if object has the model stored
-  if (!is.na(object$raw)) {
-
-    # Create temporary model for the model loading
-    object2 <- lgb.load(model_str = object$raw)
-
-    # Restore best iteration and recorded evaluations
-    object2$best_iter <- object$best_iter
-    object2$record_evals <- object$record_evals
-    object2$params <- object$params
-
-    # Return newly loaded object
-    return(object2)
-
-  } else {
-
-    # Return RDS loaded object
-    return(object)
-
-  }
-
+  lgb.restore_handle(object)
+  return(object)
 }
