@@ -109,12 +109,13 @@ class Log {
   }
   static void Fatal(const char *format, ...) {
     va_list val;
-    char str_buf[1024];
+    const size_t kBufSize = 1024;
+    char str_buf[kBufSize];
     va_start(val, format);
 #ifdef _MSC_VER
-    vsprintf_s(str_buf, format, val);
+    vsnprintf_s(str_buf, kBufSize, format, val);
 #else
-    vsprintf(str_buf, format, val);
+    vsnprintf(str_buf, kBufSize, format, val);
 #endif
     va_end(val);
 
